#!/usr/bin/env python

"""
Train an agent on Sonic using PPO2 from OpenAI Baselines.
"""

import tensorflow as tf
import numpy as np
import time

import baselines.ppo2.ppo2 as ppo2
from baselines.common.vec_env.subproc_vec_env import SubprocVecEnv
from baselines.common.vec_env.vec_monitor import VecMonitor
from baselines.common.vec_env.dummy_vec_env import DummyVecEnv
from baselines.common.vec_env.vec_env import VecEnv
from baselines import logger

from agents.sonic_util import create_envs, make_env

def train_ppo(save_path, optimizer):
	# Configure logger
	logger.configure(save_path + "/logs")

	t0 = time.time()

	save_file = None
	env_fns, env_names = create_envs()

	config = tf.ConfigProto()
	config.gpu_options.allow_growth = True
	with tf.Session(config=config, graph=tf.Graph()):
		env = VecMonitor(SubprocVecEnv(env_fns), filename=save_path + "/monitor.csv")
		model = ppo2.learn(network='cnn',
							env=env,
							optimizer=optimizer,
							nsteps=1024,
							nminibatches=8,
							lam=0.95,
							gamma=0.99,
							noptepochs=3,
							log_interval=1,
							save_interval=10,
							ent_coef=0.01,
							lr=lambda _: 2e-4,
							cliprange=lambda _: 0.1,
							total_timesteps=48e6)	
		model.save(save_path + "/model")

	t1 = time.time()
	print("Train PPO execution time: " + str(t1 - t0))
	return save_file


def evaluate_ppo(load_path, optimizer):
	# Configure logger
	logger.configure(load_path + "/logs")

	t0 = time.time()

	env_fns, env_names = create_envs(training=False)

	config = tf.ConfigProto()
	config.gpu_options.allow_growth = True

	i = 0
	for env_fn, env_name in zip(env_fns,env_names):
		i += 1
		with tf.Session(config=config, graph=tf.Graph()):
			env = VecMonitor(SubprocVecEnv([env_fn]*48), filename=load_path + "/level_" + str(i) + "_monitor.csv")
			model = ppo2.learn(network='cnn',
							env=env,
							optimizer=optimizer,
							nsteps=1024,
							nminibatches=8,
							lam=0.95,
							gamma=0.99,
							noptepochs=3,
							log_interval=1,
							ent_coef=0.01,
							lr=lambda _: 2e-4,
							cliprange=lambda _: 0.1,
							total_timesteps=1e6,
							load_path=load_path + "/model")
			env.close()

			# Create an environment an run an episode
			level = env_name.split("/")
			print(level)
			env = make_env(game=level[0], state=level[1], bk2dir=load_path)
			obs = env.reset()
			done = False
			while not done:
				action, _, _, _ = model.step([obs]*48)
				obs, rew, done, _ = env.step(action[0])
			env.close()

	t1 = time.time()
	print("Evaluate PPO execution time: " + str(t1 - t0))