#!/usr/bin/env python

"""
Train an agent on Sonic using an open source Rainbow DQN
implementation.
"""

import tensorflow as tf
import time
import csv

from anyrl.algos.dqn import DQN
from anyrl.envs.gym import BatchedGymEnv
from anyrl.envs.wrappers.batched import BatchedFrameStack
from anyrl.models.dqn_dist import rainbow_models
from anyrl.rollouts.players import BatchedPlayer, NStepPlayer
from anyrl.rollouts.replay import PrioritizedReplayBuffer
from anyrl.spaces.gym import gym_space_vectorizer
from baselines import logger

from agents.sonic_util import AllowBacktracking, create_envs, make_env

def train_rainbow(save_path, optimizer):
	open(save_path + "/model", "a").close()
	return

def evaluate_rainbow(load_path, optimizer):
	# Configure logger
	logger.configure(load_path + "/logs")

	t0 = time.time()

	env_fns, env_names = create_envs(training=False, stack=False, scale_rew=False)

	config = tf.ConfigProto()
	config.gpu_options.allow_growth = True

	i = 0
	for env_fn, env_name in zip(env_fns,env_names):
		i += 1

		env = AllowBacktracking(env_fn())
		env = BatchedFrameStack(BatchedGymEnv([[env]]), num_images=4, concat=False)
		config = tf.ConfigProto()
		config.gpu_options.allow_growth = True # pylint: disable=E1101
		with tf.Session(config=config, graph=tf.Graph()) as sess:
			dqn = DQN(*rainbow_models(sess,
									  env.action_space.n,
									  gym_space_vectorizer(env.observation_space),
									  min_val=-200,
									  max_val=200))
			player = NStepPlayer(BatchedPlayer(env, dqn.online_net), 4)
			optimize = dqn.optimize(optimizer=optimizer, learning_rate=1e-4)
			sess.run(tf.global_variables_initializer())

			total_steps = 0
			monitor = []
			def _handle_ep(steps, rew, time_elapsed):
				nonlocal total_steps
				total_steps += steps
				# Append epoch to monitor logger
				ep_element = {'r': rew, 'l': steps, 't': time_elapsed}
				monitor.append(ep_element)
				return ep_element

			dqn.train(num_steps=1e6,
					  player=player,
					  replay_buffer=PrioritizedReplayBuffer(500000, 0.5, 0.4, epsilon=0.1),
					  optimize_op=optimize,
					  train_interval=1,
					  target_interval=1024,
					  batch_size=16,
					  min_buffer_size=20000,
					  handle_ep=_handle_ep)

			env.close()

			# Save monitor logs
			save_monitor_logs(monitor, load_path + "/level_" + str(i) + "_monitor.csv")

			# Create an environment an run an episode
			level = env_name.split("/")
			print(level)
			env = make_env(game=level[0], state=level[1], stack=False, scale_rew=False, bk2dir=load_path)
			obs = env.reset()
			done = False
			while not done:
				action = dqn.step([[obs]*4], None)
				obs, rew, done, _ = env.step(action["actions"][0])
			env.close()

	t1 = time.time()
	print("Evaluate Rainbow execution time: " + str(t1 - t0))

def save_monitor_logs(epinfos, save_file):
	with open(save_file, mode='w') as csv_file:
		fieldnames = ['r', 'l', 't']
		writer = csv.DictWriter(csv_file, fieldnames=fieldnames)

		writer.writeheader()
		for epinfo in epinfos:
			writer.writerow(epinfo)


