import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import argparse
import os
import time

def main():
	args = parse_args()
	if args.ppo_dir is None or args.rainbow_dir is None:
		print("Both directories are required.")
		return

	ppo_dir = args.ppo_dir
	rainbow_dir = args.rainbow_dir
	# Append '/' if not set
	if not ppo_dir.endswith("/"):
		ppo_dir += "/"
	if not rainbow_dir.endswith("/"):
		rainbow_dir += "/"

	# The directory contains trainings with each optimitzer, so for each of them get last evaluation directory
	experiment_mapping = {}
	for element in os.listdir(ppo_dir):
		train_folder = ppo_dir + element
		if os.path.isdir(train_folder) and element.startswith("ppo_adam"):
			for elem in reversed(os.listdir(train_folder)):
				eval_folder = train_folder + '/' + elem
				if os.path.isdir(eval_folder) and elem.startswith('evaluation'):
					experiment_mapping["ppo"] = eval_folder
					break
	for element in os.listdir(rainbow_dir):
		train_folder = rainbow_dir + element
		if os.path.isdir(train_folder) and element.startswith("rainbow_adam"):
			for elem in reversed(os.listdir(train_folder)):
				eval_folder = train_folder + '/' + elem
				if os.path.isdir(eval_folder) and elem.startswith('evaluation'):
					experiment_mapping["rainbow"] = eval_folder
					break

	experiments = {}
	# Foreach optimizer compute metrics
	for drl_alg, folder in experiment_mapping.items():
		# Read the csv from evaluation phase for each level
		if drl_alg == "ppo":
			levels_monitor = read_ppo_monitor_files(folder)
			levels_progress = read_ppo_progress_files(folder)
			scale = 100
		else:
			levels_monitor = read_rainbow_monitor_files(folder)
			levels_progress = read_rainbow_progress_files(folder)
			scale = 1
		# Get learning curve values to plot
		lc_x, lc_y = get_learning_curve_values(levels_progress, scale)

		# Get loss evolution values to plot
		le_x, le_y = get_loss_evolution_values(levels_progress)

		# Construct experiment metrics
		experiment = {
			'learning_curve_values': [lc_x, lc_y], 
			'loss_values': [le_x, le_y],
		}

		# Append to experiments as {optimizer => experiment}
		experiments[drl_alg] = experiment
	# Plot
	current_unix_time = int(time.time())
	save_dir = ppo_dir + 'results_' + str(current_unix_time)
	os.mkdir(save_dir)
	plot_and_save_learning_curves(experiments, save_dir)
	plot_and_save_loss_evolution(experiments, save_dir)

def plot_and_save_learning_curves(experiments, save_dir):
	plt.xlabel('Nombre de passos')
	plt.ylabel('Puntuació')
	#plt.title('Corbes d\'aprenentatge del Joint PPO')
	for optim, experiment in experiments.items():
		plt.plot(experiment['learning_curve_values'][0], experiment['learning_curve_values'][1], label=optim)
	plt.legend()
	plt.savefig(save_dir + '/learning_curve.png')
	plt.clf()
	print('Saving learning curve into ' + save_dir + '/learning_curve.png')

def plot_and_save_loss_evolution(experiments, save_dir):
	plt.xlabel('Nombre de passos')
	plt.ylabel('Valor de la funció de pèrdua')
	#plt.title('Evolució de la funció de pèrdua del Joint PPO')
	for optim, experiment in experiments.items():
		plt.plot(experiment['loss_values'][0], experiment['loss_values'][1], label=optim)
	plt.legend()
	plt.savefig(save_dir + '/loss_evolution.png')
	print('Saving loss evolution into ' + save_dir + '/loss_evolution.png')

def get_loss_evolution_values(levels_progress):
	losses = {}
	for i in range(len(levels_progress)):
		losses['loss' + str(i)] = levels_progress[i]['loss/total']

	joint_losses = pd.DataFrame(losses)
	return levels_progress[0]['misc/total_timesteps'].values, joint_losses.mean(axis=1).values

def get_learning_curve_values(levels_progress, scale):
	eprewmeans = {}
	for i in range(len(levels_progress)):
		# Reescale ppo rewards (on experiments are scaled by 0.01)
		eprewmeans['eprewmean' + str(i)] = levels_progress[i]['eprewmean'] * scale

	joint_eprewmean = pd.DataFrame(eprewmeans)
	return levels_progress[0]['misc/total_timesteps'].values, joint_eprewmean.mean(axis=1).values

def read_ppo_monitor_files(directory):
	df1 = pd.read_csv(directory + "/level_1_monitor.csv", skiprows=1)
	df2 = pd.read_csv(directory + "/level_2_monitor.csv", skiprows=1)
	df3 = pd.read_csv(directory + "/level_3_monitor.csv", skiprows=1)
	df4 = pd.read_csv(directory + "/level_4_monitor.csv", skiprows=1)
	df5 = pd.read_csv(directory + "/level_5_monitor.csv", skiprows=1)
	df6 = pd.read_csv(directory + "/level_6_monitor.csv", skiprows=1)
	df7 = pd.read_csv(directory + "/level_7_monitor.csv", skiprows=1)
	df8 = pd.read_csv(directory + "/level_8_monitor.csv", skiprows=1)
	df9 = pd.read_csv(directory + "/level_9_monitor.csv", skiprows=1)
	df10 = pd.read_csv(directory + "/level_10_monitor.csv", skiprows=1)
	return [df1, df2, df3, df4, df5, df6, df7, df8, df9, df10]

def read_ppo_progress_files(directory):
	df1 = pd.read_csv(directory + "/level_1_logs/progress.csv")
	df2 = pd.read_csv(directory + "/level_2_logs/progress.csv")
	df3 = pd.read_csv(directory + "/level_3_logs/progress.csv")
	df4 = pd.read_csv(directory + "/level_4_logs/progress.csv")
	df5 = pd.read_csv(directory + "/level_5_logs/progress.csv")
	df6 = pd.read_csv(directory + "/level_6_logs/progress.csv")
	df7 = pd.read_csv(directory + "/level_7_logs/progress.csv")
	df8 = pd.read_csv(directory + "/level_8_logs/progress.csv")
	df9 = pd.read_csv(directory + "/level_9_logs/progress.csv")
	df10 = pd.read_csv(directory + "/level_10_logs/progress.csv")
	return [df1, df2, df3, df4, df5, df6, df7, df8, df9, df10]

def read_rainbow_monitor_files(directory):
	df1 = pd.read_csv(directory + "/level_1_monitor.csv")
	df2 = pd.read_csv(directory + "/level_2_monitor.csv")
	df3 = pd.read_csv(directory + "/level_3_monitor.csv")
	df4 = pd.read_csv(directory + "/level_4_monitor.csv")
	df5 = pd.read_csv(directory + "/level_5_monitor.csv")
	df6 = pd.read_csv(directory + "/level_6_monitor.csv")
	df7 = pd.read_csv(directory + "/level_7_monitor.csv")
	df8 = pd.read_csv(directory + "/level_8_monitor.csv")
	df9 = pd.read_csv(directory + "/level_9_monitor.csv")
	df10 = pd.read_csv(directory + "/level_10_monitor.csv")
	return [df1, df2, df3, df4, df5, df6, df7, df8, df9, df10]

def read_rainbow_progress_files(directory):
	df1 = pd.read_csv(directory + "/level_1_logs/progress.csv")
	df2 = pd.read_csv(directory + "/level_2_logs/progress.csv")
	df3 = pd.read_csv(directory + "/level_3_logs/progress.csv")
	df4 = pd.read_csv(directory + "/level_4_logs/progress.csv")
	df5 = pd.read_csv(directory + "/level_5_logs/progress.csv")
	df6 = pd.read_csv(directory + "/level_6_logs/progress.csv")
	df7 = pd.read_csv(directory + "/level_7_logs/progress.csv")
	df8 = pd.read_csv(directory + "/level_8_logs/progress.csv")
	df9 = pd.read_csv(directory + "/level_9_logs/progress.csv")
	df10 = pd.read_csv(directory + "/level_10_logs/progress.csv")
	return [df1, df2, df3, df4, df5, df6, df7, df8, df9, df10]

def parse_args():
	parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
	parser.add_argument('--ppo_dir', help='Directory containing ppo executions', type=str, default=None)
	parser.add_argument('--rainbow_dir', help='Directory containing rainbow executions', type=str, default=None)
	return parser.parse_args()

if __name__ == '__main__':
	main()