import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import argparse
import os
import time

optimizers_mapping = {
	'adam': "Adam",
	'adagrad': "AdaGrad",
	'adadelta': "Adadelta",
	'nag': "NAG",
	'momentum': "Momentum",
}

def main():
	args = parse_args()
	if args.dir is None:
		print("The directory argument is required.")
		return

	base_directory = args.dir
	# Append '/' if not set
	if not base_directory.endswith("/"):
		base_directory += "/"
	# The directory contains trainings with each optimitzer, so for each of them get last evaluation directory
	experiment_mapping = {}
	for element in os.listdir(base_directory):
		train_folder = base_directory + element
		if os.path.isdir(train_folder) and element.startswith('rainbow'):
			for elem in reversed(os.listdir(train_folder)):
				eval_folder = train_folder + '/' + elem
				if os.path.isdir(eval_folder) and elem.startswith('evaluation'):
					optim = element.split("_")[1]
					experiment_mapping[optim] = eval_folder
					break

	experiments = {}
	# Foreach optimizer compute metrics
	for optim, folder in experiment_mapping.items():
		# Read the csv from evaluation phase for each level
		levels_monitor = read_monitor_files(folder)
		levels_progress = read_progress_files(folder)

		# Compute mean score metric
		mean_score = compute_mean_score(levels_monitor)

		# Get learning curve values to plot
		lc_x, lc_y = get_learning_curve_values(levels_progress)

		# Get mean total time elapsed
		mean_time_elapsed = compute_mean_time_elapsed(levels_progress)

		# Get mean time elapsed per update
		mean_time_elapsed_per_update = compute_mean_time_elapsed_per_update(levels_progress)

		# Get loss evolution values to plot
		le_x, le_y = get_loss_evolution_values(levels_progress)

		# Construct experiment metrics
		experiment = {
			'mean_score': mean_score, 
			'learning_curve_values': [lc_x, lc_y], 
			'mean_time_elapsed': mean_time_elapsed, 
			'mean_time_elapsed_per_update': mean_time_elapsed_per_update,
			'loss_values': [le_x, le_y],
		}

		# Append to experiments as {optimizer => experiment}
		experiments[optim] = experiment
	# Plot
	current_unix_time = int(time.time())
	save_dir = base_directory + 'results_' + str(current_unix_time)
	os.mkdir(save_dir)
	plot_and_save_learning_curves(experiments, save_dir)
	plot_and_save_loss_evolution(experiments, save_dir)
	save_results(experiments, save_dir)

def plot_and_save_learning_curves(experiments, save_dir):
	plt.xlabel('Nombre de passos')
	plt.ylabel('Puntuació')
	plt.title('Corbes d\'aprenentatge del Rainbow')
	for optim, experiment in experiments.items():
		plt.plot(experiment['learning_curve_values'][0], experiment['learning_curve_values'][1], label=optim)
	plt.legend()
	plt.savefig(save_dir + '/learning_curve.png')
	plt.clf()
	print('Saving learning curve into ' + save_dir + '/learning_curve.png')

def plot_and_save_loss_evolution(experiments, save_dir):
	plt.xlabel('Nombre de passos')
	plt.ylabel('Valor de la funció de pèrdua')
	plt.title('Evolució de la funció de pèrdua del Rainbow')
	for optim, experiment in experiments.items():
		plt.plot(experiment['loss_values'][0], experiment['loss_values'][1], label=optim)
	plt.legend()
	plt.savefig(save_dir + '/loss_evolution.png')
	print('Saving loss evolution into ' + save_dir + '/loss_evolution.png')

def save_results(experiments, save_dir):
	with open(save_dir + '/metrics_table.txt', 'w') as f:
		f.write('\\begin{center}\n')
		f.write('\t\\begin{tabular}{|c|c|c|c|c|c|}\n')

		optims = '\t\t'
		scores = '\t\tPuntuació'
		total_time = '\t\tTemps per nivell'
		time_per_update = '\t\tTemps per Actualització'

		for optim, experiment in experiments.items():
			optim = optimizers_mapping[optim]
			optims += ' & ' + optim
			scores += ' & ' + str(experiment['mean_score'])
			total_time += ' & ' + str(experiment['mean_time_elapsed'])
			time_per_update += ' & ' + str(experiment['mean_time_elapsed_per_update'])
		optims += ' \\\\\n'
		scores += ' \\\\\n'
		total_time += ' \\\\\n'
		time_per_update += ' \\\\\n'
		f.write('\t\t\\hline\n')
		f.write(optims)
		f.write('\t\t\\hline\n')
		f.write(scores)
		f.write('\t\t\\hline\n')
		f.write(total_time)
		f.write('\t\t\\hline\n')
		f.write(time_per_update)
		f.write('\t\t\\hline\n')
		f.write('\t\\end{tabular}\n')
		f.write('\\end{center}')
	print('Saving escalar results into ' + save_dir + '/metrics_table.txt')

def get_loss_evolution_values(levels_progress):
	losses = {}
	for i in range(len(levels_progress)):
		losses['loss' + str(i)] = levels_progress[i]['loss/total']

	joint_losses = pd.DataFrame(losses)
	return levels_progress[0]['misc/total_timesteps'].values, joint_losses.mean(axis=1).values

def compute_mean_time_elapsed_per_update(levels_progress):
	mean_sum = 0
	# Sum the mean of each episode
	for level_progress in levels_progress:
		mean_sum += level_progress['misc/time_elapsed_on_params_update'].mean()
	return round(mean_sum / len(levels_progress), 4)

def compute_mean_time_elapsed(levels_progress):
	mean_sum = 0
	for level_progress in levels_progress:
		mean_sum += level_progress['misc/time_elapsed'].iloc[-1]
	return round(mean_sum / len(levels_progress), 2)

def get_learning_curve_values(levels_progress):
	eprewmeans = {}
	for i in range(len(levels_progress)):
		eprewmeans['eprewmean' + str(i)] = levels_progress[i]['eprewmean']

	joint_eprewmean = pd.DataFrame(eprewmeans)
	return levels_progress[0]['misc/total_timesteps'].values, joint_eprewmean.mean(axis=1).values

def compute_mean_score(levels_monitor):
	mean_sum = 0
	# Sum the mean of each episode
	for level_monitor in levels_monitor:
		mean_sum += level_monitor['r'].mean()
	return round((mean_sum) / len(levels_monitor), 2)

def read_monitor_files(directory):
	df1 = pd.read_csv(directory + "/level_1_monitor.csv")
	df2 = pd.read_csv(directory + "/level_2_monitor.csv")
	df3 = pd.read_csv(directory + "/level_3_monitor.csv")
	df4 = pd.read_csv(directory + "/level_4_monitor.csv")
	df5 = pd.read_csv(directory + "/level_5_monitor.csv")
	df6 = pd.read_csv(directory + "/level_6_monitor.csv")
	df7 = pd.read_csv(directory + "/level_7_monitor.csv")
	df8 = pd.read_csv(directory + "/level_8_monitor.csv")
	df9 = pd.read_csv(directory + "/level_9_monitor.csv")
	df10 = pd.read_csv(directory + "/level_10_monitor.csv")
	return [df1, df2, df3, df4, df5, df6, df7, df8, df9, df10]

def read_progress_files(directory):
	df1 = pd.read_csv(directory + "/level_1_logs/progress.csv")
	df2 = pd.read_csv(directory + "/level_2_logs/progress.csv")
	df3 = pd.read_csv(directory + "/level_3_logs/progress.csv")
	df4 = pd.read_csv(directory + "/level_4_logs/progress.csv")
	df5 = pd.read_csv(directory + "/level_5_logs/progress.csv")
	df6 = pd.read_csv(directory + "/level_6_logs/progress.csv")
	df7 = pd.read_csv(directory + "/level_7_logs/progress.csv")
	df8 = pd.read_csv(directory + "/level_8_logs/progress.csv")
	df9 = pd.read_csv(directory + "/level_9_logs/progress.csv")
	df10 = pd.read_csv(directory + "/level_10_logs/progress.csv")
	return [df1, df2, df3, df4, df5, df6, df7, df8, df9, df10]

def parse_args():
	parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
	parser.add_argument('--dir', help='Directory containing all executions', type=str, default=None)
	return parser.parse_args()

if __name__ == '__main__':
	main()