#!/usr/bin/env python
import warnings
import argparse
import time
import os
import shutil

from agents.ppo import train_ppo, evaluate_ppo
from agents.rainbow import train_rainbow, evaluate_rainbow

def main():
	# Parse arguments
	args = parse_args()

	if args.alg != "ppo" and args.alg != "rainbow":
		print("Algorithm must be one of: 'ppo' 'rainbow'.")
		print("Exiting...")
		return

	if args.optim != "adam" and args.optim != "adadelta" and args.optim != "adagrad" and args.optim != "nag" and args.optim != "momentum":
		print("Algorithm must be one of: 'adam' 'adadelta' 'adagrad' 'nag' 'momentum'.")
		print("Exiting...")
		return

	current_unix_time = int(time.time())

	path = build_directory_struct(args, current_unix_time)
	if path is None:
		return
	print("Logging into %s." % path)

	# Build log path
	log_path = path + "/logs"
	os.mkdir(log_path)

	# Run Training / Testing
	if not args.test:
		if args.alg == 'ppo':
			train_ppo(path, args.optim)
		elif args.alg == 'rainbow':
			train_rainbow(path, args.optim)
		else:
			print("Invalid algorithm: %s" % args.alg)
			return
	else:
		if args.alg == 'ppo':
			model = evaluate_ppo(path, args.optim)
		elif args.alg == 'rainbow':
			model = evaluate_rainbow(path, args.optim)
		else:
			print("Invalid algorithm: %s" % args.alg)
			return

def build_directory_struct(args, current_unix_time):
	path = None
	# Build the path for training/testing
	if not args.test:
		# Check if save path has been specified
		if args.save_path is None:
			print("The save_path is required.")
			print("Exiting...")
			return None
		# Set the path as the save path
		path = args.save_path

		# If directory does not exist, create it
		if not os.path.isdir(path):
			os.mkdir(path)

		# Append '/' if not set
		if not path.endswith("/"):
			path += "/"

		# Create a directory inside the path to identify the current run
		path += args.alg + "_" + args.optim + "_training_" + str(current_unix_time)
		os.mkdir(path)
	else:
		# Check if load path has been specified
		if args.load_path is None:
			print("The load_path is required.")
			print("Exiting...")
			return
		# Set the path as the load path
		path = args.load_path

		# If directory does not exist, exit
		if not os.path.isdir(path):
			print("The load_path is not a valid directory.")
			print("Exiting...")
			return

		# Append '/' if not set
		if not path.endswith("/"):
			path += "/"

		# Build the model path to load
		model_path = path + "model"
		# If model file does not exists, exit
		if not os.path.isfile(model_path):
			print("Unable to find model.")
			print("Exiting...")
			return None

		# Create a directory inside the path for evaluation logging
		path += "evaluation_" + str(current_unix_time)
		os.mkdir(path)
		# Copy saved model to evaluation folder
		shutil.copyfile(model_path, path + "/model")

	return path


def parse_args():
	parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

	parser.add_argument('--alg', help='Algorithm', type=str, default='ppo')
	parser.add_argument('--optim', help='Optimizer', type=str, default='adam')
	parser.add_argument('--save_path', help='Path to save trained model to', type=str, default=None)
	parser.add_argument('--load_path', help='Path to load trained model to', type=str, default=None)
	parser.add_argument('--test', default=False, action="store_true")

	return parser.parse_args()

if __name__ == '__main__':
	with warnings.catch_warnings():
		warnings.filterwarnings("ignore")
		main()
